use crate::{local_storage, setLocation, Issue, PartialIssue, User};
use sycamore::prelude::*;

#[component(IssueCreate<G>)]
pub fn issue_create() -> View<G> {
    let repo = Signal::new(String::new());
    let title = Signal::new(String::new());
    let desc = Signal::new(String::new());

    let error = Signal::new(false);

    let current_user = Signal::new(User::default());

    if G::IS_BROWSER {
        if let Some(user) = local_storage::getItem("token") {
            let decoded = base64::decode(user).unwrap();
            let json = unsafe { String::from_utf8_unchecked(decoded) };
            current_user.set(serde_json::from_str(&json).unwrap());
        } else {
            setLocation("/login");
        }
    }

    let on_click = cloned!((repo, title, desc, error) => move |_| {
        let repo = (*repo.get()).trim().to_string();
        let author = current_user.get().id;
        let title = (*title.get()).trim().to_string();
        let desc = (*desc.get()).trim().to_string();

        if repo == "" || title == "" || desc == "" {
            error.set(true);
            return;
        }

        let issue = PartialIssue {
            repo,
            author,
            title,
            desc
        };

        let sered = serde_json::to_string(&issue).unwrap();

        wasm_bindgen_futures::spawn_local(async move {
            let resp = reqwasm::http::Request::post("/api/add_issue")
                .body(sered)
                .header("Content-Type", "application/json")
                .send()
                .await
                .unwrap()
                .text()
                .await
                .unwrap();

            let formatted = format!("/issue/{resp}");
            setLocation(&formatted);
        });
    });

    view! {
        div(class="card") {
            a(href="/") { button { "<<<" } }
            label { "repo:"
                // TODO: Fetch a list of repos from gitea or something
                input(placeholder="karx/bugspray", bind:value=repo)
            }
            label {
                "Title:"
                input(bind:value=title)
            }
            label {
                "Description:"
                textarea(bind:value=desc)
            }
            (if *error.clone().get() {
                view! {
                    p(style="color: red") { "Invalid form input. Make sure none of the inputs are empty." }
                }
            } else {
                view! {}
            })
            div(style="justify-content: right") {
                button(on:click=on_click) { "Submit" }
            }
        }
    }
}

#[component(IssueDetail<G>)]
pub fn issue_detail(index: i64) -> View<G> {
    let issue: Signal<Issue> = Signal::new(Issue::default());
    let user: Signal<User> = Signal::new(User::default());
    if G::IS_BROWSER {
        wasm_bindgen_futures::spawn_local(cloned!((issue, user) => async move {
            let resp = reqwasm::http::Request::get(&format!("/api/issues/{}", index)).send().await.unwrap();
            if resp.status() == 404 {
                setLocation("/not_found"); // lmao
            }
            let text = resp.text().await.unwrap();
            let parsed: (Issue, User) = serde_json::from_str(&text).unwrap();
            issue.set(parsed.0);
            user.set(parsed.1);
        }));
    }

    let title = create_memo(cloned!(issue => move || issue.get().title.clone()));
    // let author = create_memo(cloned!(issue => move || issue.get().author.clone()));
    let author = create_memo(cloned!(user => move || user.get().username.clone()));
    let repo = create_memo(cloned!(issue => move || issue.get().repo.clone()));
    let desc = create_memo(cloned!(issue => move || issue.get().desc.clone()));

    view! {
        div(class="card") {
            h3(class="text-align-center", style="margin-bottom: 0") { (title.get()) }
            div(style="justify-content: space-between; display: flex") {
                div { (author.get()) }
                div { (repo.get()) }
            }
            br {}
            br {}
            p(class="text-align-center") { (desc.get()) }
            div(style="display: flex; justify-content: left") {
                a(href="/") { button { "<<<" } }
            }
        }
    }
}
