use crate::{local_storage, log, setLocation, PartialUser, User};
use base64::encode;
use fancy_regex::Regex;
use pbkdf2::password_hash::PasswordVerifier;
use sycamore::prelude::*;

#[component(Register<G>)]
pub fn register() -> View<G> {
    let username = Signal::new(String::new());
    let password_raw = Signal::new(String::new());
    let password_confirm = Signal::new(String::new());

    let error_blank = Signal::new(false);
    let error_passmatch = Signal::new(false);

    let error_username_format = Signal::new(false);
    let error_format = Signal::new(false);

    let error_unique_username = Signal::new(false);

    let whitespace_re = Regex::new(r"\s").unwrap();
    let password_re = Regex::new(
        r"^.*(?=.{8})(?=.+[a-z])(?=.*[A-Z])(?=.+\d)(?=.*[`~!@#$%^&*()_+\-=\[\]{}\\|:;/?<>,.]).*$",
    )
    .unwrap();

    let on_click = cloned!((error_blank, error_passmatch, username, password_raw, password_confirm, error_format, error_username_format, error_unique_username) => move |_| {
        use pbkdf2::{
            password_hash::{
                PasswordHasher, SaltString,
            },
            Pbkdf2
        };
        use rand_core::OsRng;

        let username = username.get().trim().to_string();
        let password_raw = password_raw.get().trim().to_string();
        let password_confirm = password_confirm.get().trim().to_string();

        error_blank.set(username == "" || password_raw == "" || password_confirm == "");
        error_passmatch.set(password_raw != password_confirm);
        error_format.set(!password_re.is_match(&password_raw).unwrap());
        error_username_format.set(whitespace_re.is_match(&username).unwrap()); // error is true if username contains whitespace

        wasm_bindgen_futures::spawn_local(cloned!((error_unique_username, username) => async move {
            let resp = reqwasm::http::Request::get(&format!("/api/user/by_username/{username}"))
                .send()
                .await
                .unwrap()
                .text()
                .await
                .unwrap();

            let desered: User = serde_json::from_str(&resp).unwrap();

            error_unique_username.set(desered.id != 0);
        }));

        if *error_unique_username.get() || *error_blank.get() || *error_passmatch.get() || *error_format.get() || *error_username_format.get() {
            return;
        }

        let salt = SaltString::generate(&mut OsRng);
        let password_hash = Pbkdf2.hash_password(password_raw.as_bytes(), &salt).unwrap().to_string();

        log(&password_hash);

        let user = PartialUser {
            username,
            password_hash
        };

        let serialized = serde_json::to_string(&user).unwrap();

        wasm_bindgen_futures::spawn_local(async move {
            reqwasm::http::Request::post("/api/register")
                .body(serialized)
                .header("Content-Type", "application/json")
                .send()
                .await
                .unwrap()
                .text()
                .await
                .unwrap();

            setLocation("/login");
        });
    });

    view! {
        div(class="card") {
            label { "Username:"
                input(bind:value=username)
            }
            label { "Password"
                input(bind:value=password_raw, type="password")
            }
            label { "Confirm password:"
                input(bind:value=password_confirm, type="password")
            }

            div {
                (if *error_unique_username.get() {
                    view! {
                        p(style="color: red") { "That username is not available." }
                    }
                } else {
                    view! {}
                })
            }

            div {
                (if *error_username_format.get() {
                    view! {
                        p(style="color: red") { "Username cannot contain any spaces." }
                    }
                } else {
                    view! {}
                })
            }

            div {
                (if *error_format.get() {
                    view! {
                        p(style="color: red") { "Invalid password. Make sure it contains lowercase and uppercase letters, at least one digit, and at least one special character." }
                    }
                } else {
                    view! {}
                })
            }

            div {
                (if *error_passmatch.get() {
                    view! {
                        p(style="color: red") { "Passwords do not match" }
                    }
                } else {
                    view! {}
                })
            }

            div {
                (if *error_blank.get() {
                    view! {
                        p(style="color: red") { "Invalid form input. Make sure none of the fields are empty." }
                    }
                } else {
                    view! {}
                })
            }

            div(style="justify-content: right") {
                button(on:click=on_click) { "Register" }
            }

            small {
                a(href="/login") { "Already have an account?" }
            }
        }
    }
}

#[component(Login<G>)]
pub fn login() -> View<G> {
    let username = Signal::new(String::new());
    let password_raw = Signal::new(String::new());

    let error_blank = Signal::new(false);
    let error_wrong_username = Signal::new(false);
    let error_wrong_password = Signal::new(false);

    let text = Signal::new(String::from("Log in"));

    let on_click = cloned!((username, password_raw, error_blank, error_wrong_password, error_wrong_username, text) => move |_| {
        let username = username.get().trim().to_string();
        let password_raw = password_raw.get().trim().to_string();

        error_blank.set(username == "" || password_raw == "");

        wasm_bindgen_futures::spawn_local(cloned!((error_wrong_username, error_wrong_password, text) => async move {
            use pbkdf2::{
                password_hash::PasswordHash,
                Pbkdf2,
            };

            text.set(String::from("Logging in..."));

            let resp = reqwasm::http::Request
                ::get(&format!("/api/user/by_username/{username}"))
                .send()
                .await
                .unwrap()
                .text()
                .await
                .unwrap();

            let user: User = serde_json::from_str(&resp).unwrap();

            error_wrong_username.set(user.id == 0);

            if *error_wrong_username.get() {
                text.set(String::from("Log in"));
                return;
            }

            let hashed = PasswordHash::new(&user.password_hash).unwrap();

            error_wrong_password.set(Pbkdf2.verify_password(password_raw.as_bytes(), &hashed).is_err());

            if *error_wrong_password.get() {
                text.set(String::from("Log in"));
                return;
            }

            let encoded = encode(serde_json::to_string(&user).unwrap().as_bytes());

            local_storage::setItem("token", &encoded);
            log(&encoded);

            setLocation("/");

            text.set(String::from("Log in"));
        }));
    });

    view! {
        div(class="card") {
            label { "Username:"
                input(bind:value=username)
            }
            label { "Password:"
                input(bind:value=password_raw, type="password")
            }

            div {
                (if *error_blank.get() {
                    view! {
                        p(style="color: red") { "Invalid form input. Make sure none of the fields are blank." }
                    }
                } else {
                    view! {}
                })
            }

            div {
                (if *error_wrong_username.get() {
                    view! {
                        p(style="color: red") { "No such user exists!" }
                    }
                } else {
                    view! {}
                })
            }

            div {
                (if *error_wrong_password.get() {
                    view! {
                        p(style="color: red") { "Incorrect password." }
                    }
                } else {
                    view! {}
                })
            }

            div(style="justify-content: right") {
                button(on:click=on_click) { (text.get()) }
            }

            small {
                a(href="/register") { "Don't have an account?" }
            }
        }
    }
}

#[component(Logout<G>)]
pub fn logout() -> View<G> {
    let on_confirm = |_| {
        local_storage::removeItem("token");
        setLocation("/");
    };

    let on_cancel = |_| {
        setLocation("/");
    };

    view! {
        div(class="card") {
            div(class="text-align-center") {
                "Really log out?"
                br {}
                button(on:click=on_confirm) { "Yes, log out" }
                button(on:click=on_cancel) { "Cancel" }
            }
        }
    }
}
