mod components;
mod local_storage;

use serde::{Deserialize, Serialize};
use sycamore::prelude::*;
use sycamore_router::{
    HistoryIntegration, Route, Router, RouterProps, StaticRouter, StaticRouterProps,
};
use wasm_bindgen::prelude::*;

macro_rules! wasm_import {
    ($($tt:tt)*) => {
        #[wasm_bindgen]
        extern "C" {
            #[wasm_bindgen]
            pub fn $($tt)*;
        }
    };
}

#[macro_export]
macro_rules! wasm_import_with_ns {
    ($ns: ident, $($tt:tt)*) => {
        #[wasm_bindgen]
        extern "C" {
            #[wasm_bindgen(js_namespace = $ns)]
            pub fn $($tt)*;
        }
    };
}

wasm_import!(setLocation(href: &str));
wasm_import_with_ns!(console, log(s: &str));

#[derive(PartialEq, Eq, Debug, Default, Clone, Serialize, Deserialize)]
pub struct Issue {
    pub id: i64,
    pub repo: String,
    pub author: i32,
    pub title: String,
    pub desc: String,
    pub timestamp: u64,
}

#[derive(PartialEq, Eq, Debug, Default, Clone, Serialize, Deserialize)]
pub struct PartialIssue {
    pub repo: String,
    pub author: i32,
    pub title: String,
    pub desc: String,
}

#[derive(PartialEq, Eq, Debug, Default, Clone, Serialize, Deserialize)]
pub struct PartialUser {
    pub username: String,
    pub password_hash: String,
}

#[derive(PartialEq, Eq, Debug, Default, Clone, Serialize, Deserialize)]
pub struct User {
    pub id: i32,
    pub username: String,
    pub password_hash: String,
}

#[derive(Route, Clone)]
pub enum AppRoutes {
    #[to("/")]
    Index,
    #[to("/issue/<index>")]
    IssueDetail(i64),
    #[to("/add_issue")]
    IssueCreate,

    #[to("/register")]
    Register,
    #[to("/login")]
    Login,
    #[to("/logout")]
    Logout,

    #[not_found]
    NotFound,
}

fn switch<G: Html>(route: ReadSignal<AppRoutes>) -> View<G> {
    let logged_in = if G::IS_BROWSER {
        local_storage::getItem("token").is_some()
    } else {
        false
    };
    view! {
        div(class="wrapper") {
            (match route.get().as_ref() {
                AppRoutes::Index => view! {
                    h1(class="text-align-center") { "Bugspray" }
                    (if !logged_in {
                        view! {
                            a(href="/register") { "Register" }
                            " | "
                            a(href="/login") { "Log in" }
                        }
                    } else {
                        view! {
                            a(href="/add_issue") { "+ New Issue" }
                            " | "
                            a(href="/logout") { "Log out" }
                        }
                    })
                    components::home::HomeScreen()
                },
                AppRoutes::IssueDetail(index) => view! {
                    h1(class="text-align-center") { "Bugspray" }
                    (if !logged_in {
                        view! {
                            a(href="/register") { "Register" }
                            " | "
                            a(href="/login") { "Log in" }
                        }
                    } else {
                        view! {
                            a(href="/add_issue") { "+ New Issue" }
                            " | "
                            a(href="/logout") { "Log out" }
                        }
                    })
                    components::issues::IssueDetail(*index)
                },
                AppRoutes::IssueCreate => view! {
                    h1(class="text-align-center") { "Bugspray" }
                    components::issues::IssueCreate()
                },

                AppRoutes::Register => view! {
                    h1(class="text-align-center") { "Bugspray" }
                    components::users::Register()
                },
                AppRoutes::Login => view! {
                    h1(class="text-align-center") { "Bugspray" }
                    components::users::Login()
                },
                AppRoutes::Logout => view! {
                    h1(class="text-align-center") { "Bugspray" }
                    components::users::Logout()
                },

                _ => view! {
                    h1(class="text-align-center") { "Bugspray" }
                    (if !logged_in {
                        view! {
                            a(href="/register") { "Register" }
                            " | "
                            a(href="/login") { "Log in" }
                            br {}
                        }
                    } else {
                        view! {
                            a(href="/add_issue") { "+ New Issue" }
                            " | "
                            a(href="/logout") { "Log out" }
                        }
                    })
                    "404 Not Found"
                },
            })
        }
    }
}

#[component(App<G>)]
pub fn app(path: Option<String>) -> View<G> {
    match path {
        Some(pathname) => {
            let route = AppRoutes::match_path(&pathname);
            view! {
                StaticRouter(StaticRouterProps::new(route, |route: AppRoutes| switch(Signal::new(route).handle())))
            }
        }
        None => view! {
            Router(RouterProps::new(HistoryIntegration::new(), switch))
        },
    }
}
