use rusqlite::{params, OptionalExtension};
use serde::{Deserialize, Serialize};
use tokio::sync::{
    mpsc::{unbounded_channel, UnboundedReceiver, UnboundedSender},
    oneshot,
};

#[derive(Debug)]
enum Task {
    // channel, id
    GetIssue(oneshot::Sender<Option<Issue>>, i64),
    // channel, repo, author, title, desc
    AddIssue(oneshot::Sender<Option<i64>>, String, i32, String, String),
    // channel
    AllIssues(oneshot::Sender<Vec<Issue>>),

    // channel(id), username, hashed password
    CreateUser(oneshot::Sender<Option<i64>>, String, String),

    // channel(user), username
    GetUserByUsername(oneshot::Sender<Option<User>>, String),
    GetUserById(oneshot::Sender<Option<User>>, i32),
}

pub struct DbExecutor {
    rx: UnboundedReceiver<Task>,
    db: rusqlite::Connection,
}

#[derive(Clone)]
pub struct ExecutorConnection {
    tx: UnboundedSender<Task>,
}

#[derive(Debug, Clone, Serialize, Deserialize)]
pub struct Issue {
    pub id: i64,
    pub repo: String,
    pub author: i32,
    pub title: String,
    pub desc: String,
    pub timestamp: u64,
}

#[derive(Debug, Clone, Serialize, Deserialize)]
pub struct PartialIssue {
    pub repo: String,
    pub author: i32,
    pub title: String,
    pub desc: String,
}

#[derive(Debug, Clone, Serialize, Deserialize)]
pub struct PartialUser {
    pub username: String,
    pub password_hash: String,
}

#[derive(Debug, Clone, Serialize, Deserialize, Default)]
pub struct User {
    pub id: i32,
    pub username: String,
    pub password_hash: String,
}

macro_rules! executor_wrapper {
    ($name:ident, $task:expr, $ret:ty, $($arg:ident: $ty:ty),*) => {
        pub async fn $name(&self, $($arg: $ty),*) -> $ret {
            let (otx, orx) = oneshot::channel();
            self.tx.send($task(otx, $($arg),*)).unwrap();
            orx.await.unwrap()
        }
    };
    ($name:ident, $task:expr, $ret:ty) => {
        pub async fn $name(&self) -> $ret {
            let (otx, orx) = oneshot::channel();
            self.tx.send($task(otx)).unwrap();
            orx.await.unwrap()
        }
    }
}

impl ExecutorConnection {
    executor_wrapper!(
        add_issue,
        Task::AddIssue,
        Option<i64>,
        repo: String,
        author: i32,
        title: String,
        desc: String
    );
    executor_wrapper!(get_issue, Task::GetIssue, Option<Issue>, id: i64);
    executor_wrapper!(all_issues, Task::AllIssues, Vec<Issue>);

    executor_wrapper!(
        create_user,
        Task::CreateUser,
        Option<i64>,
        username: String,
        password_hash: String
    );
    executor_wrapper!(
        get_user_by_username,
        Task::GetUserByUsername,
        Option<User>,
        username: String
    );
    executor_wrapper!(get_user_by_id, Task::GetUserById, Option<User>, id: i32);
}

impl DbExecutor {
    pub fn create(dbpath: &str) -> rusqlite::Result<(Self, ExecutorConnection)> {
        let (tx, rx) = unbounded_channel();

        let db = rusqlite::Connection::open(dbpath)?;
        db.execute("create table if not exists users(id integer primary key, username text not null, password_hash text not null)", [])?;
        db.execute("create table if not exists issues(id integer primary key, repo text not null, author integer not null, title text not null, desc text not null, timestamp integer default (strftime('%s','now')), foreign key(author) references users(id))", [])?;

        Ok((Self { rx, db }, ExecutorConnection { tx }))
    }

    pub fn run(&mut self) {
        while let Some(task) = self.rx.blocking_recv() {
            match task {
                Task::GetIssue(tx, id) => {
                    let issue = self
                        .db
                        .query_row("select * from issues where id=?", params![id], |r| {
                            Ok(Issue {
                                id: r.get(0).unwrap_or_default(),
                                repo: r.get(1).unwrap_or_default(),
                                author: r.get(2).unwrap_or_default(),
                                title: r.get(3).unwrap_or_default(),
                                desc: r.get(4).unwrap_or_default(),
                                timestamp: r.get(5).unwrap_or_default(),
                            })
                        })
                        .optional()
                        .unwrap();
                    tx.send(issue).unwrap();
                }
                Task::AddIssue(tx, repo, author, title, desc) => {
                    let id = self.db.query_row("insert into issues(repo, author, title, desc) values(?, ?, ?, ?) returning id", params![repo, author, title, desc], |r| r.get(0)).optional().unwrap_or_default();
                    tx.send(id).unwrap();
                }
                Task::AllIssues(tx) => {
                    let mut stmt = self.db.prepare("select * from issues;").unwrap();
                    let mut rows = stmt.query([]).unwrap();

                    let mut issues = Vec::new();
                    while let Some(r) = rows.next().unwrap() {
                        issues.push(Issue {
                            id: r.get(0).unwrap_or_default(),
                            repo: r.get(1).unwrap_or_default(),
                            author: r.get(2).unwrap_or_default(),
                            title: r.get(3).unwrap_or_default(),
                            desc: r.get(4).unwrap_or_default(),
                            timestamp: r.get(5).unwrap_or_default(),
                        });
                    }

                    tx.send(issues).unwrap();
                }
                Task::CreateUser(tx, username, password_hash) => {
                    // hashing will be done in the frontend so as not to leak passwords over the web

                    let id = self
                        .db
                        .query_row(
                            "insert into users(username, password_hash) values(?, ?) returning id",
                            params![username, password_hash],
                            |r| r.get(0),
                        )
                        .optional()
                        .unwrap();
                    tx.send(id).unwrap();
                }
                Task::GetUserByUsername(tx, username) => {
                    let user = self
                        .db
                        .query_row(
                            "select * from users where username=?",
                            params![username],
                            |r| {
                                Ok(User {
                                    id: r.get(0).unwrap(),
                                    username: r.get(1).unwrap(),
                                    password_hash: r.get(2).unwrap(),
                                })
                            },
                        )
                        .optional()
                        .unwrap();
                    tx.send(user).unwrap();
                }
                Task::GetUserById(tx, id) => {
                    let user = self
                        .db
                        .query_row("select * from users where id=?", params![id], |r| {
                            Ok(User {
                                id: r.get(0).unwrap(),
                                username: r.get(1).unwrap(),
                                password_hash: r.get(2).unwrap(),
                            })
                        })
                        .optional()
                        .unwrap();

                    tx.send(user).unwrap();
                }
            }
        }
    }
}
